use std::{
    collections::BTreeMap,
    time::{SystemTime, UNIX_EPOCH},
};

use anyhow::Context;
use async_trait::async_trait;
use aws_config::{BehaviorVersion, Region};
use aws_sdk_s3::{
    config::Credentials,
    primitives::{ByteStream, SdkBody},
};
use tokio::{
    io::{AsyncReadExt, BufReader},
    sync::RwLock,
};

use super::StorageBackend;

pub struct S3StorageBackend {
    client: aws_sdk_s3::Client,
    bucket: String,

    index_lock: RwLock<BTreeMap<String, RwLock<()>>>,
}

impl S3StorageBackend {
    pub async fn upload_file(&self, path: &str, buffer: &[u8]) -> anyhow::Result<()> {
        tracing::trace!("committing file: {}", &path);

        self.client
            .put_object()
            .bucket(&self.bucket)
            .key(path)
            .body(ByteStream::new(SdkBody::from(buffer)))
            .send()
            .await?;

        Ok(())
    }

    pub async fn get_file(&self, path: &str) -> anyhow::Result<Option<Vec<u8>>> {
        tracing::trace!("getting file: {}", path);

        let obj = match self
            .client
            .get_object()
            .bucket(&self.bucket)
            .key(path)
            .send()
            .await
        {
            Ok(ok) => ok,
            Err(err) => match err.into_service_error() {
                aws_sdk_s3::operation::get_object::GetObjectError::NoSuchKey(_) => return Ok(None),
                e => anyhow::bail!(e.to_string()),
            },
        };

        let mut buf_reader = BufReader::new(obj.body.into_async_read());

        let mut output = Vec::new();
        buf_reader.read_buf(&mut output).await?;

        Ok(Some(output))
    }

    pub async fn append_file(&self, path: &str, buffer: &[u8]) -> anyhow::Result<()> {
        tracing::trace!("appending file: {}", &path);
        {
            let mut index_lock = self.index_lock.write().await;
            let item = index_lock.get(path);
            if item.is_none() {
                index_lock.insert(path.to_string(), RwLock::default());
            }
        }
        let index_lock = self.index_lock.read().await;
        let item = index_lock.get(path).expect("to find a path lock");
        let lock = item.write().await;

        let file = self.get_file(path).await?;
        match file {
            Some(mut file_contents) => {
                file_contents.extend_from_slice(buffer);
                self.upload_file(path, &file_contents).await?
            }
            None => self.upload_file(path, buffer).await?,
        }

        drop(lock);

        Ok(())
    }
}

impl S3StorageBackend {
    pub async fn new(
        key_id: impl Into<String>,
        key: impl Into<String>,
        endpint_url: impl Into<String>,
        bucket: impl Into<String>,
    ) -> anyhow::Result<Self> {
        let shared_config = aws_config::defaults(BehaviorVersion::latest())
            .region(Region::new("eu-west-1"))
            .credentials_provider(Credentials::new(
                key_id,
                key,
                None,
                None,
                env!("CARGO_PKG_NAME"),
            ));

        let config = aws_sdk_s3::config::Builder::from(&shared_config.load().await)
            .endpoint_url(endpint_url)
            .force_path_style(true)
            .build();

        let client = aws_sdk_s3::Client::from_conf(config);

        Ok(Self {
            client,
            bucket: bucket.into(),
            index_lock: RwLock::default(),
        })
    }

    pub async fn new_from_env() -> anyhow::Result<Self> {
        let key_id = std::env::var("AWS_ACCESS_KEY_ID").context("AWS_ACCESS_KEY_ID was not set")?;
        let access_key =
            std::env::var("AWS_SECRET_ACCESS_KEY").context("AWS_SECRET_ACCESS_KEY was not set")?;
        let endpoint_url =
            std::env::var("AWS_ENDPOINT_URL").context("AWS_ENDPOINT_URL was not set")?;
        let bucket = std::env::var("AWS_BUCKET").context("AWS_BUCKET was not set")?;

        Self::new(key_id, access_key, endpoint_url, bucket).await
    }
}

#[async_trait]
impl StorageBackend for S3StorageBackend {
    async fn flush_segment(&self, topic: &str, buffer: &[u8]) -> anyhow::Result<String> {
        let segment_key = uuid::Uuid::now_v7();

        self.upload_file(
            &format!("nodata/logs/{}/{}.pb", topic, &segment_key.to_string()),
            buffer,
        )
        .await?;

        Ok(segment_key.to_string())
    }
    async fn append_index(
        &self,
        topic: &str,
        segment_file: &str,
        time: SystemTime,
    ) -> anyhow::Result<()> {
        self.append_file(
            &format!("nodata/indexes/{}", topic),
            format!(
                "{},{}\n",
                time.duration_since(UNIX_EPOCH)
                    .expect("to be able to get time")
                    .as_secs(),
                segment_file
            )
            .as_bytes(),
        )
        .await
    }
}
